from io import BytesIO
from .stream import BinStream
from enum import IntEnum
import gzip
import pyzstd
from xxhash import xxh64, xxh3_64

signature_to_extension = {
    b'OggS': 'ogg',
    bytes.fromhex('00010000'): 'ttf',
    bytes.fromhex('1a45dfa3'): 'webm',
    b'true': 'ttf',
    b'OTTO\0': 'otf',
    b'"use strict";': 'min.js',
    b'<template ': 'template.html',
    b'<!-- Elements -->': 'template.html',
    b'DDS ': 'dds',
    b'<svg': 'svg',
    b'PROP': 'bin',
    b'PTCH': 'bin',
    b'BKHD': 'bnk',
    b'r3d2Mesh': 'scb',
    b'r3d2anmd': 'anm',
    b'r3d2canm': 'anm',
    b'r3d2sklt': 'skl',
    b'r3d2': 'wpk',
    bytes.fromhex('33221100'): 'skn',
    b'PreLoadBuildingBlocks = {': 'preload',
    b'\x1bLuaQ\x00\x01\x04\x04': 'luabin',
    b'\x1bLuaQ\x00\x01\x04\x08': 'luabin64',
    bytes.fromhex('023d0028'): 'troybin',
    b'[ObjectBegin]': 'sco',
    b'OEGM': 'mapgeo',
    b'TEX\0': 'tex',
    b'RW': 'wad'
}


def guess_extension(data):
    if data[4:8] == bytes.fromhex('c34ffd22'):
        return 'skl'
    else:
        for signature, extension in signature_to_extension.items():
            if data.startswith(signature):
                return extension


def parse_extension(path):
    if path.endswith('.wad.client'):
        return 'wad'
    for _, extension in signature_to_extension.items():
        if path.endswith(extension):
            return extension


def hash_to_hex(hash):
    return f'{hash:016x}'


def hex_to_hash(hex):
    return int(hex, 16)


def name_to_hash(name):
    return xxh64(name.lower()).intdigest()


def hex_to_name(hashtables, table_name, hash):
    return hashtables.get(table_name, {}).get(hash, hash)


def name_to_hex(name):
    return xxh64(name.lower()).hexdigest()


def name_or_hex_to_hash(value):
    try:
        return hex_to_hash(value)
    except:
        return name_to_hash(value)


class WADCompressionType(IntEnum):
    Raw = 0
    Gzip = 1
    Satellite = 2
    Zstd = 3
    ZstdChunked = 4

    def __json__(self):
        return self.name


class WADChunk:
    __slots__ = (
        'id', 'hash', 'offset',
        'compressed_size', 'decompressed_size', 'compression_type',
        'duplicated', 'subchunk_start', 'subchunk_count',
        'checksum', 'data', 'extension'
    )

    def __init__(self, id=None, hash=None, offset=None, compressed_size=None, decompressed_size=None, compression_type=None, duplicated=None, subchunk_start=None, subchunk_count=None, checksum=None, data=None, extension=None):
        self.id = id
        self.hash = hash
        self.offset = offset
        self.compressed_size = compressed_size
        self.decompressed_size = decompressed_size
        self.compression_type = compression_type
        self.duplicated = duplicated
        self.subchunk_start = subchunk_start
        self.subchunk_count = subchunk_count
        self.checksum = checksum
        self.data = data
        self.extension = extension

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__ if key != 'data'}

    @staticmethod
    def default(*, id=0, hash='', offset=0, compressed_size=0, decompressed_size=0, compression_type=WADCompressionType.Raw, duplicated=False, subchunk_start=0, subchunk_count=0, checksum=0):
        chunk = WADChunk()
        chunk.id = id
        chunk.hash = hash
        chunk.offset = offset
        chunk.compressed_size = compressed_size
        chunk.decompressed_size = decompressed_size
        chunk.compression_type = compression_type
        chunk.duplicated = duplicated
        chunk.subchunk_start = subchunk_start
        chunk.subchunk_count = subchunk_count
        chunk.checksum = checksum
        return chunk

    def free_data(self):
        self.data = None

    def read_data(self, bs):
        # read data and decompress
        bs.seek(self.offset)
        raw = bs.read(self.compressed_size)
        if self.compression_type == WADCompressionType.Raw:
            self.data = raw
        elif self.compression_type == WADCompressionType.Gzip:
            self.data = gzip.decompress(raw)
        elif self.compression_type == WADCompressionType.Satellite:
            # Satellite is not supported
            self.data = None
        elif self.compression_type == WADCompressionType.Zstd:
            self.data = pyzstd.decompress(raw)
        elif self.compression_type == WADCompressionType.ZstdChunked:
            if raw[:4] == b'\x28\xb5\x2f\xfd':
                self.data = pyzstd.decompress(raw)
            else:
                self.data = raw
        # guess extension
        if self.extension == None:
            self.extension = guess_extension(self.data)

    def write_data(self, bs, chunk_id, chunk_hash, chunk_data, *, previous_chunks=None):
        self.hash = chunk_hash
        if self.extension in ('bnk', 'wpk'):
            self.data = chunk_data
            self.compression_type = WADCompressionType.Raw
        else:
            self.data = pyzstd.compress(chunk_data)
            self.compression_type = WADCompressionType.Zstd
        self.compressed_size = len(self.data)
        self.decompressed_size = len(chunk_data)
        self.checksum = xxh3_64(self.data).intdigest()
        # check duplicated data
        if previous_chunks:
            duped_id, duped_chunk = None, None
            for id, chunk in enumerate(previous_chunks):
                if chunk.checksum == self.checksum and chunk.compressed_size == self.compressed_size and chunk.decompressed_size == self.decompressed_size:
                    duped_id = id
                    duped_chunk = chunk
                    break
            if duped_chunk != None:
                # if there is a duped chunk in previous
                if not duped_chunk.duplicated:
                    # if the chunk was not a duped chunk
                    # rewrite the duplicated value for the previous chunk
                    duped_chunk.duplicated = True
                    bs.seek(272 + duped_id * 32 + 21)
                    bs.write_b(duped_chunk.duplicated)
                # set this chunk as duplicated and copy the offset from duped chunk
                self.duplicated = True
                self.offset = duped_chunk.offset
        if not self.duplicated:
            # if its duplicated dont need to write data
            # go to end file, save data offset and write chunk data
            bs.seek(0, 2)
            self.offset = bs.tell()
            bs.write(self.data)
        # go to this chunk offset and write stuffs
        # hack: the first chunk start at 272 (because we write version 3.3)
        self.id = chunk_id
        chunk_offset = 272 + chunk_id * 32
        bs.seek(chunk_offset)
        bs.write_u64(name_or_hex_to_hash(chunk_hash))
        bs.write_u32(
            self.offset,
            self.compressed_size,
            self.decompressed_size
        )
        bs.write_u8(self.compression_type.value)
        bs.write_b(self.duplicated)
        bs.write_u16(0)
        bs.write_u64(self.checksum)


class WAD:
    __slots__ = ('signature', 'version', 'chunks')

    def __init__(self, signature=None, version=None, chunks=None):
        self.signature = signature
        self.version = version
        self.chunks = chunks

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__ if key != 'IO'}

    def stream(self, path, mode, raw=None):
        if raw != None:
            if raw == True:  # the bool True value
                return BinStream(BytesIO())
            else:
                return BinStream(BytesIO(raw))
        return BinStream(open(path, mode))

    def read(self, path, raw=None):
        with self.stream(path, 'rb', raw) as bs:
            # read header
            self.signature, = bs.read_s(2)
            if self.signature != 'RW':
                raise Exception(
                    f'pyRitoFile: Failed: Read WAD {path}: Wrong file signature: {self.signature}')
            major, minor = bs.read_u8(2)
            self.version = float(f'{major}.{minor}')
            if major > 3:
                raise Exception(
                    f'pyRitoFile: Failed: Read WAD {path}: Unsupported file version: {self.version}')
            wad_checksum = 0
            if major == 2:
                ecdsa_len = bs.read_u8()
                bs.pad(83)
                wad_checksum, = bs.read_u64()
            elif major == 3:
                bs.pad(256)
                wad_checksum, = bs.read_u64()
            if major == 1 or major == 2:
                toc_start_offset, toc_file_entry_size = bs.read_u16(
                    2)
            # read chunks
            chunk_count, = bs.read_u32()
            self.chunks = [WADChunk() for i in range(chunk_count)]
            for chunk_id, chunk in enumerate(self.chunks):
                chunk.id = chunk_id
                chunk.hash = hash_to_hex(bs.read_u64()[0])
                chunk.offset, chunk.compressed_size, chunk.decompressed_size, = bs.read_u32(
                    3)
                chunk.compression_type = WADCompressionType(
                    bs.read_u8()[0] & 15)
                chunk.duplicated, = bs.read_b()
                chunk.subchunk_start, = bs.read_u16()
                chunk.subchunk_count = chunk.compression_type.value >> 4
                chunk.checksum = bs.read_u64()[0] if major >= 2 else 0

    def write(self, path, raw=None):
        with self.stream(path, 'wb', raw) as bs:
            # write header
            bs.write_s('RW')  # signature
            bs.write_u8(3, 3)  # version
            bs.write(b'\x00' * 256)  # pad 256 bytes
            bs.write_u64(0)  # wad checksum
            bs.write_u32(len(self.chunks))
            # write chunks
            for chunk in self.chunks:
                bs.write_u64(name_or_hex_to_hash(chunk.hash))
                bs.write_u32(
                    chunk.offset,
                    chunk.compressed_size,
                    chunk.decompressed_size
                )
                bs.write_u8(chunk.compression_type.value)
                bs.write_b(chunk.duplicated)
                bs.write_u16(chunk.subchunk_start)
                bs.write_u64(chunk.checksum)
            return bs.raw() if raw else None

    def un_hash(self, hashtables=None):
        if hashtables == None:
            return
        for chunk in self.chunks:
            chunk.hash = hex_to_name(hashtables, 'hashes.game.txt', chunk.hash)
            if '.' in chunk.hash and chunk.extension == None:
                chunk.extension = parse_extension(chunk.hash)
        self.chunks = sorted(self.chunks, key=lambda chunk: chunk.hash)



from io import BytesIO
from typing import Optional, Tuple
from .structs import BoundingBox, BoundingSphere, Vector
from .stream import BinStream
from .ermmm import FNV1a
from enum import IntEnum

def bin_hash(name):
    return f'{FNV1a(name):08x}'


class SKNVertexType(IntEnum):
    BASIC = 0
    COLOR = 1
    TANGENT = 2

    def __json__(self):
        return self.name


class SKNVertex:
    """
    Represents a single Vertex.
    
    Fields:
    --------
        `Optionals`:\n
            Based in :class:`SKN` Vertex_Type.\n
            `color`: A tuple, that represents an (RGBA), vertex_type in (SKNVertexType.COLOR, SKNVertexType.TANGENT).\n
            `tangent`: :class:`Vector`, vertex_type == SKNVertexType.TANGENT.\n

        `Persistents between every vertex type`:\n
            `position`: :class:`Vector`, local XYZ position.\n
            `normal`: :class:`Vector`, normals of the vertex.
            `influences`: A tuple, representing each unique bone id that has influence in the vertex.\n
            `weights`: A tuple, representing how much weight the certain bone id has in the vertex.\n
            # Example of reading influences and weights:\n
            # Assume influence (102, 55, 0, 0), and weights (0.65, 0.35, 0, 0).\n
            # The joint with ID of 102 have an weight of 0.65 in the vertex.\n
            # The joint with ID of 55 have an weight of 0.35 in the vertex.\n
            # Both has to be with lenght 4, since league only allows max 4 influences in a bound skin mesh.\n

    Methods:
    -------
        `__json__()`: Returns a dict of every fields. (field: value of the field)
    """
    __slots__ = (
        'position', 'influences', 'weights', 'normal', 'uv',
        'color', 'tangent'
    )

    def __init__(self):
        self.position: Optional[Vector] = None
        self.influences: Optional[Tuple[int, int, int, int]] = None
        self.weights: Optional[Tuple[float, float, float, float]] = None
        self.normal: Optional[Vector] = None
        self.uv: Optional[Vector] = None
        self.color: Optional[Tuple[int, int, int, int]] = None
        self.tangent: Optional[Vector] = None

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}


class SKNSubmesh:
    """
    Represents a single Submesh.

    Fields:
    --------
        `name`: String, name of the submesh.\n
        `bin_hash`: String, name of the submesh as FNV1a lowered hash.\n
        `vertex_start`: Integer, where the vertex start.\n
        `vertex_count`: Integer, amount of unique vertices.\n
        `index_start`: Integer, where the index starts.\n
        `index_count`: Integer, amount of unique indices.\n

    Methods:
    -------
        `__json__()`: Returns a dict of every fields. (field: value of the field)
    """
    __slots__ = (
        'name', 'bin_hash',
        'vertex_start', 'vertex_count', 'index_start', 'index_count'
    )

    def __init__(self):
        self.name: Optional[str] = None
        self.bin_hash: Optional[str] = None
        self.vertex_start: Optional[int] = None
        self.vertex_count: Optional[int] = None
        self.index_start: Optional[int] = None
        self.index_count: Optional[int] = None

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}


class SKN:
    """
    Represents a Simple Skin Mesh (3D File) from League of Legends, use .read() to get the fields initialized.

    Fields:
    -------
        Optionals for version with major being 4 (4.0, 4.1, ...):\n
            `bounding_box`: A :class:`BoundingBox` that represents the local bounding box, it has: min_vec, max_vec; while min and max together creates a Cube.\n
            `bounding_sphere`: A :class:`BoundingSphere` that represents the local bounding sphere, it has: point, radius; while point and radius together creates a Sphere.\n
            `flags`: Integer that represents the flags.\n
            `vertex_type`: :class:`SKNVertexType` that represents the vertex type.\n
            `vertex_size`: Integer that represents the vertex size.\n
        
        Persistents between every version:
            `signature`: A string that contains the hex value of the SKN signature.\n
            `version`: A float that represents the version (major.minor).\n
            `submeshes`: A list of :class:`SKNSubmesh`.\n
            `vertices`: A list of :class:`SKNVertex`.\n
            `indices`: A list of integers that represents every 3 vertice index thats connects to a triangle,
                example: (0, 5, 4, ...) vertex 0, 5, 4 connects together to make a triangle.\n
    
    Methods:
    -------
        `read()`: Used to read any supported SKN file, it fills the fields of the instance.\n
        `__json__()`: Returns a dict of every fields. (field: value of the field).\n
        `stream()`: Used to manually write / read SKN files.
    """
    __slots__ = (
        'signature', 'version', 'flags',
        'bounding_box', 'bounding_sphere', 'vertex_type', 'vertex_size',
        'submeshes', 'indices', 'vertices'
    )
    
    def __init__(self):
        self.signature: Optional[int] = None
        self.version: Optional[float] = None
        self.flags: Optional[int] = None
        self.bounding_box: Optional[BoundingBox] = None
        self.bounding_sphere: Optional[BoundingSphere] = None
        self.vertex_type: Optional[int] = None
        self.vertex_size: Optional[int] = None
        self.submeshes: list[SKNSubmesh] = []
        self.indices: list[int] = []
        self.vertices: list[SKNVertex] = []

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}
    
    def stream(self, path, mode, raw=None):
        if raw != None:
            if raw == True:  # the bool True value
                return BinStream(BytesIO())
            else:
                return BinStream(BytesIO(raw))
        return BinStream(open(path, mode))

    def read(self, path: str, raw: Optional[bytes] = None):
        """
        Initialize the fields of :class:`SKN`

        Parameters
        ------------
        `path`: Optional[:class:`str`]
            File path to read the SKN using an existing file that should work in `open(path)`.\n
        
        `raw`: Optional[:class:`bytes`]
            None (`Default value`): Recommended if reading from an file in local hard drive.
            bytes: Reads SKN information from a bytes object.\n

        Raises
        --------
        `WrongFileSignature.`\n
        `UnsupportedFileVersion.`\n
        `InvalidMethodUsage.`\n
        """
        with self.stream(path, 'rb', raw) as bs:
            self.signature, = bs.read_u32()
            if self.signature != 0x00112233:
                raise Exception(
                    f'pyRitoFile: Failed: Read SKN {path}: Wrong signature file: {hex(self.signature)}')
            self.signature = hex(self.signature)

            major, minor = bs.read_u16(2)
            self.version = float(f'{major}.{minor}')
            if major not in (0, 2, 4) and minor != 1:
                raise Exception(
                    f'pyRitoFile: Failed: Read SKN {path}: Unsupported file version: {major}.{minor}')
            
            if major == 0:
                # version 0 doesn't have submesh data
                index_count, vertex_count = bs.read_u32(2)

                submesh = SKNSubmesh()
                submesh.name = 'Base'
                submesh.bin_hash = bin_hash(submesh.name)
                submesh.vertex_start = 0
                submesh.vertex_count = vertex_count
                submesh.index_start = 0
                submesh.index_count = index_count
                self.submeshes = [submesh]
            else:
                # read submeshes
                submesh_count, = bs.read_u32()
                self.submeshes = [SKNSubmesh() for i in range(submesh_count)]
                for submesh in self.submeshes:
                    submesh.name, = bs.read_s_padded(64)
                    submesh.bin_hash = bin_hash(submesh.name)
                    submesh.vertex_start, submesh.vertex_count, submesh.index_start, submesh.index_count = bs.read_u32(
                        4)

                if major == 4:
                    self.flags, = bs.read_u32()

                index_count, vertex_count = bs.read_u32(2)
                if major == 4:
                    self.vertex_size, = bs.read_u32()
                    self.vertex_type, = bs.read_u32()
                    if not self.vertex_type in (0, 1, 2):
                        raise Exception(f'pyRitoFile: Failed: Read SKN {path}: Invalid vertex_type: {self.vertex_type}')
                    self.vertex_type = SKNVertexType(self.vertex_type)
                    self.bounding_box = BoundingBox(bs.read_vec3()[0], bs.read_vec3()[0])
                    self.bounding_sphere = BoundingSphere(
                        bs.read_vec3()[0], bs.read_f32()[0])
                    
            if index_count % 3 > 0:
                raise Exception(f'pyRitoFile: Failed: Read SKN {path}: Bad indices data: {index_count}')

            # read unique indices
            indices = bs.read_u16(index_count)
            self.indices = []
            for i in range(0, index_count, 3):
                if indices[i] == indices[i+1] or indices[i+1] == indices[i+2] or indices[i+2] == indices[i]:
                    continue
                self.indices.extend(
                    (indices[i], indices[i+1], indices[i+2]))

            # read vertices
            self.vertices = [SKNVertex() for i in range(vertex_count)]
            for vertex in self.vertices:
                vertex.position, = bs.read_vec3()
                vertex.influences = bs.read_u8(4)
                vertex.weights = bs.read_f32(4)
                vertex.normal, = bs.read_vec3()
                vertex.uv, = bs.read_vec2()
                if self.vertex_type in (SKNVertexType.COLOR, SKNVertexType.TANGENT):
                    vertex.color = bs.read_u8(4)
                    if self.vertex_type == SKNVertexType.TANGENT:
                        vertex.tangent, = bs.read_vec4()

    def write(self, path, raw=None):
        """
        Exports an bytes with :class:`SKN` infos

        Parameters
        ------------
        `path`: Optional[:class:`str`]
            File path to write the SKN that should work in `open(path, 'wb')`.\n
        
        `raw`: Optional[:class:`bytes`]
            None (`Default value`): Recommended if writing in local memory.
            bytes: Outputs the SKN information in an bytes object.\n

        """
        with self.stream(path, 'wb', raw) as bs:
            # magic, version
            bs.write_u32(0x00112233)
            if self.version != None and self.version >= 4:
                bs.write_u16(4, 1)
            else:
                bs.write_u16(1, 1)
                self.version = 1.1
            # submesh
            bs.write_u32(len(self.submeshes))
            for submesh in self.submeshes:
                bs.write_s_padded(submesh.name, 64)
                bs.write_u32(
                    submesh.vertex_start, submesh.vertex_count, submesh.index_start, submesh.index_count)
            if self.version >= 4:
                bs.write_u32(self.flags)
            # stuffs
            bs.write_u32(len(self.indices), len(self.vertices))
            if self.version >= 4:
                bs.write_u32(self.vertex_size, self.vertex_type.value)
                bs.write_vec3(*self.bounding_box)
                bs.write_vec3(self.bounding_sphere[0])
                bs.write_f32(self.bounding_sphere[1])
             # indices vertices
            bs.write_u16(*self.indices)
            for vertex in self.vertices:
                bs.write_vec3(vertex.position)
                bs.write_u8(*vertex.influences)
                bs.write_f32(*vertex.weights)
                bs.write_vec3(vertex.normal)
                bs.write_vec2(vertex.uv)
                if self.vertex_type in (SKNVertexType.COLOR, SKNVertexType.TANGENT):
                    bs.write_u8(*vertex.color)
                    if self.vertex_type == SKNVertexType.TANGENT:
                        bs.write_vec4(vertex.tangent)

            return bs.raw() if raw else None



from .skl import SKL, SKLJoint
from .skn import SKN, SKNVertex, SKNSubmesh
from .so import SO, SOFlag
from .anm import ANM, ANMPose, ANMTrack, ANMErrorMetric
from .mapgeo import MAPGEO
from .bin import BIN, BINEntry, BINPatch, BINField, BINType, BINHelper, name_to_hex as bin_hash
from .bnk import BNK, BNKObjectType
from .wpk import WPK
from .tex import TEX, TEXFormat
from .wad import WAD, WADChunk, WADCompressionType, name_to_hex as wad_hash, guess_extension
from .ermmm import Elf, FNV1, FNV1a
from .rst import RST
from . import stream
from json import dump, dumps, JSONEncoder

__version__ = "0.2.4"

class CACHED_BIN_HASHES(dict):
    def __getitem__(self, key):
        if key in self.keys():
            return super().__getitem__(key)
        else:
            super().__setitem__(key, bin_hash(key))
            return super().__getitem__(key)


class PRFEncoder(JSONEncoder):
    def default(self, obj):
        if hasattr(obj, '__json__'):
            return obj.__json__()
        elif isinstance(obj, bytes):
            return str(obj.hex(' ').upper())
        else:
            return JSONEncoder.default(self, obj)


def write_json(path, obj):
    good_types = [SKL, SKN, SO, ANM, MAPGEO, BIN, BNK, WPK, TEX, WAD]
    with open(path, 'w+') as f:
        for good_type in good_types:
            if isinstance(obj, good_type):
                dump(obj, f, indent=4, cls=PRFEncoder)


def to_json(obj):
    good_types = [SKL, SKN, SO, ANM, MAPGEO, BIN, BNK, WPK, TEX, WAD]
    for good_type in good_types:
        if isinstance(obj, good_type):
            return dumps(obj, indent=4, cls=PRFEncoder)
        

def read_skl(path, raw=None):
    skl = SKL()
    skl.read(path, raw)
    return skl


def write_skl(path, skl):
    skl.write(path)


def read_skn(path, raw=None):
    skn = SKN()
    skn.read(path, raw)
    return skn


def write_skn(path, skn):
    skn.write(path)


def read_sco(path, raw=None):
    so = SO()
    so.read_sco(path, raw)
    return so

def write_sco(path, so):
    so.write_sco(path)

def read_scb(path, raw=None):
    so = SO()
    so.read_scb(path, raw)
    return so

def write_scb(path, so):
    so.write_scb(path)


def read_anm(path, raw=None):
    anm = ANM()
    anm.read(path, raw)
    return anm

def write_anm(path, anm):
    anm.write(path)



def read_mapgeo(path, raw=None):
    mg = MAPGEO()
    mg.read(path, raw)
    return mg

def write_mapgeo(path, mg, version):
    mg.write(path, version)


def read_bin(path, raw=None):
    bin = BIN()
    bin.read(path, raw)
    return bin


def write_bin(path, bin):
    bin.write(path)


def read_bnk(path, raw=None):
    bnk = BNK()
    bnk.read(path, raw)
    return bnk

def write_bnk(path, bnk, wem_datas):
    bnk.write(path, wem_datas)

def read_wpk(path, raw=None):
    wpk = WPK()
    wpk.read(path, raw)
    return wpk

def write_wpk(path, wpk, wem_datas):
    wpk.write(path, wem_datas)


def read_tex(path, raw=None):
    tex = TEX()
    tex.read(path, raw)
    return tex


def read_wad(path, raw=None):
    wad = WAD()
    wad.read(path, raw)
    return wad


def write_wad(path, wad):
    wad.write(path)
